@@ -58,16 +58,17 @@ def _setex(conn, key, value, time):
 if list(map(int, redis.__version__.split('.'))) < [2, 4, 12]:
     raise Exception("Upgrade your Redis client to version 2.4.12 or later")
 
-VERSION = '0.31.1'
+VERSION = '0.32.1'
 
 RPQUEUE_CONFIGS = {}
 
 __all__ = ['VERSION', 'new_rpqueue', 'set_key_prefix',
     'set_redis_connection_settings', 'set_redis_connection',
-    'task', 'periodic_task', 'cron_task', 'Task', 'Data',
+    'task', 'periodic_task', 'cron_task', 'Task', 'EnqueuedTask', 'Data',
     'get_task', 'result', 'results',
     'set_priority', 'known_queues', 'queue_sizes', 'clear_queue',
-    'execute_tasks', 'SimpleLock', 'NoLock', 'script_load']
+    'execute_tasks', 'SimpleLock', 'NoLock', 'script_load',
+    'call_task', 'known_tasks']
 __all__.sort()
 
 def new_rpqueue(name, pfix=None):
@@ -126,7 +127,7 @@ _NAME_KEYS = dict(
     DATA_KEY = b'dataq:',
     DATAM_KEY = b'dataqm:',
     DATAV_KEY = b'dataqv:',
-
+    REGISTRY_KEY = b'taskregistry:',
 )
 # silence Pyflakes
 QUEUES_KNOWN = _NAME_KEYS['QUEUES_KNOWN']
@@ -144,6 +145,7 @@ VISIBILITY_KEY = _NAME_KEYS['VISIBILITY_KEY']
 DATA_KEY = _NAME_KEYS['DATA_KEY']
 DATAM_KEY = _NAME_KEYS['DATAM_KEY']
 DATAV_KEY = _NAME_KEYS['DATAV_KEY']
+REGISTRY_KEY = _NAME_KEYS['REGISTRY_KEY']
 KEY_PREFIX = b''
 
 def set_key_prefix(pfix):
@@ -160,6 +162,7 @@ def set_key_prefix(pfix):
 
 PY3 = sys.version_info >= (3, 0)
 _SB = (str, bytes) if PY3 else (str, unicode)
+REGISTRY_CHANGED = 0
 REGISTRY = {}
 NEVER_SKIP = set()
 SHOULD_QUIT = multiprocessing.Array('i', (0,), lock=False)
@@ -169,6 +172,9 @@ DEADLETTER_QUEUE = b'DEADLETTER_QUEUE'
 MINIMUM_DELAY = 1
 EXECUTE_TASKS = False
 QUEUE_ITEMS_PER_PASS = 100
+KEEP_REDIS_TASK_REGISTRY = False
+# If true, clears out the key before setting again
+REDIS_TASK_REGISTRY_AUTO_CLEAR = True
 
 REDIS_CONNECTION_SETTINGS = {}
 POOL = None
@@ -428,7 +434,7 @@ local cutoff
 
 if #ARGV == 1 then
     cutoff = redis.call('time')
-    cutoff = cutoff[1] .. '.' .. cutoff[2]
+    cutoff = string.format("%s.%06d", cutoff[1], tonumber(cutoff[2]))
 else
     cutoff = ARGV[2]
 end
@@ -524,6 +530,111 @@ def _handle_delayed(conn, queues=None, timeout=1):
 
     handle_delayed_lua_(conn, queues, timeout)
 
+def _to_str(v):
+    v = v or ''
+    if PY3:
+        return v if isinstance(v, str) else v.decode('latin-1')
+    return v if isinstance(v, unicode) else v.decode('latin-1')
+
+def _to_bytes(v):
+    v = v or b''
+    if PY3:
+        return v if isinstance(v, bytes) else v.encode('latin-1')
+    return v if isinstance(v, str) else v.encode('latin-1')
+
+def _update_redis_task_registry(last, conn=None):
+    global REGISTRY_CHANGED
+    last = last or {}
+    if not KEEP_REDIS_TASK_REGISTRY or not REGISTRY_CHANGED:
+        return last
+
+    new = {}
+    REGISTRY_CHANGED = 0
+    for name, task in REGISTRY.items():
+        queue, name, delay, attempts, retry_delay, save_results, vis_timeout, use_dead, delay
+        delay = task.delay
+        if isinstance(task.delay, CronTab):
+            delay = ' '.join(m.input for m in delay.matchers)
+        new[name] = json.dumps([_to_str(task.queue), name, delay, task.never_skip,
+            task.attempts, task.retry_delay, task.save_results,
+            task.vis_timeout, task.use_dead])
+
+    p = conn.pipeline(REDIS_TASK_REGISTRY_AUTO_CLEAR)
+    if last != new:
+        if REDIS_TASK_REGISTRY_AUTO_CLEAR:
+            p.delete(REGISTRY_KEY)
+        p.hmset(REGISTRY_KEY, new)
+        p.execute()
+
+    return new
+
+def _load_task_registry(conn=None, tasks={}):
+    tasks = tasks or conn.hgetall(REGISTRY_KEY)
+    for name, task_info in tasks.items():
+        if not isinstance(task_info, list):
+            task_info = json.loads(_to_str(task_info))
+
+        # we never have a local copy of the function
+        task_info.insert(2, None)
+        # json round-trip BS
+        task_info[0] = _to_bytes(task_info[0])
+        # But we create the task cache so we don't need to keep doing this crap
+        REGISTRY[name] = Task(*task_info)
+
+def call_task(name, queue, *args, **kwargs):
+    '''
+    Calls a task with the given name, using the provided queue if the queue was
+    otherwise not known. Returns an ``EnqueuedTask``, which you can use to track
+    task progress. Used via::
+
+        t = rpqueue.call_task('dotted.task_name', None or QUEUE, *args, **kw)
+        t.wait(1)
+        t.status
+        t.result
+
+    If you want to see a ``ValueError`` if your task is not known, pass
+    ``_verify_call=True`` as a keyword argument.
+
+    '''
+    name = _to_str(name)
+    verify = kwargs.pop("_verify_call", None)
+    if name not in REGISTRY:
+        conn = get_connection()
+        task_info = conn.hget(REGISTRY_KEY, name)
+        if not task_info and verify:
+            raise ValueError("Task name %s not known".format(name))
+        elif not task_info:
+            task_info = [queue or DEFAULT_QUEUE, name, 0, 0, 1, 0, 0, 0, 0]
+        else:
+            task_info = json.loads(task_info)
+        _load_task_registry(conn, {name: task_info})
+
+    return REGISTRY[name].execute(*args, **kwargs)
+
+class known_tasks:
+    '''
+    Read-only dict-like thing that offers access to tasks that have registeed,
+    or are known by rpqueue by enabling ``KEEP_REDIS_TASK_REGISTRY``. Returns
+    the ``Task`` to be executed. Used via::
+
+        t = rpqueue.known_tasks['dotted.task_name'].execute(*args, **kwargs)
+        t.wait(1)
+        t.status
+        t.result
+    '''
+    def __getitem__(self, name):
+        if name not in REGISTRY:
+            conn = get_connection()
+            task_info = conn.hget(REGISTRY_KEY, name)
+            if not task_info:
+                raise KeyError("Task name %s not known".format(name))
+            task_info = json.loads(task_info)
+            _load_task_registry(conn, {name: task_info})
+
+        return REGISTRY[name]
+
+known_tasks = known_tasks()
+
 def _gt(conn):
     t = conn.time()
     if not isinstance(t, (list, tuple)):
@@ -707,7 +818,7 @@ class Data(object):
         im, iv = pipe.execute()
         return (im - iv, iv)
 
-    def put_data(self, data, use_dead=None, is_one=False, chunksize=512):
+    def put_data(self, data, use_dead=None, is_one=False, chunksize=512, delay=0):
         '''
         Puts data items into the queue. Data is assumed to be a list of data
         items, each of which will be assigned a new UUID as an identifier before
@@ -717,11 +828,16 @@ class Data(object):
          - ``use_dead`` - if provided and true-ish, will spew attempts exceeded
            data items into data DEADLETTER_QUEUE (queue has a default, can be
            overridden here to enable on this inserted data)
-         - ``is_one`` - if you provide a list to ``data``, but just want that to be
-           one item, you can ensure that with ``is_one=True``
+         - ``is_one`` - if you provide a list to ``data``, but just want that to
+           be one item, you can ensure that with ``is_one=True``
          - ``chunksize`` - the number of items to insert per pipeline flush
+         - ``delay`` - if >0, will delay the inserted items from reading for the
+           provided number of seconds, requires ``vis_timeout`` > 0
 
         '''
+        delay = max(delay, 0)
+        if delay and self.vis_timeout <= 0:
+            raise ValueError("queue vis_timeout must be > 0 to delay items")
         if is_one:
             data = [data]
         use_dead = int(bool(self.use_dead if use_dead is None else use_dead))
@@ -730,12 +846,17 @@ class Data(object):
             proto['_attempts'] = self.attempts
         if use_dead:
             proto['_use_dead'] = use_dead
+        if delay:
+            proto['_attempts'] = proto.get('_attempts', 1) + 1
         proto = json.dumps(proto)
         conn = get_connection()
         t = _gt(conn)
+        if delay:
+            t2 = t + delay
         pipe = conn.pipeline(True)
         lq = DATA_KEY + self.queue
         hq = DATAM_KEY + self.queue
+        vq = DATAV_KEY + self.queue
         chunksize = max(chunksize, 1)
         added = {}
         add = {}
@@ -748,7 +869,10 @@ class Data(object):
                     pipe.rpush(lq, *list(add.values()))
                 else:
                     pipe.hmset(hq, add)
-                    pipe.rpush(lq, *list(add))
+                    if delay:
+                        _zadd(pipe, vq, {tid: t2 for tid in add})
+                    else:
+                        pipe.rpush(lq, *list(add))
                 t = _gt(pipe)
                 add = {}
         if add:
@@ -756,7 +880,10 @@ class Data(object):
                 pipe.rpush(lq, *list(add.values()))
             else:
                 pipe.hmset(hq, add)
-                pipe.rpush(lq, *list(add))
+                if delay:
+                    _zadd(pipe, vq, {tid: t2 for tid in add})
+                else:
+                    pipe.rpush(lq, *list(add))
                 pipe.execute()
 
         return added
@@ -902,7 +1029,6 @@ class Data(object):
         get_connection().delete(self.qkey, self.mkey, self.vkey)
 
 
-
 class Task(object):
     '''
     An object that represents a task to be executed. These will replace
@@ -951,6 +1077,8 @@ class Task(object):
         else:
             NEVER_SKIP.discard(name)
         REGISTRY[name] = self
+        global REGISTRY_CHANGED
+        REGISTRY_CHANGED = 1
         if self.delay:
             # periodic or cron task
             execute_delay = self.next()
@@ -1004,12 +1132,12 @@ class Task(object):
             ec = rpq._enqueue_call
             conn = rpq.get_connection()
             et = rpq._execute_task
-            enq = rpq._EnqueuedTask
+            enq = rpq.EnqueuedTask
         else:
             ec = _enqueue_call
             conn = get_connection()
             et = _execute_task
-            enq = _EnqueuedTask
+            enq = EnqueuedTask
 
         if kwargs.pop('execute_inline_now', None):
             # allow for testing/debugging to execute the task immediately
@@ -1097,9 +1225,10 @@ class _ExecutingTask(object):
         return "<ExecutingTask taskid=%s function=%s args=%r kwargs=%r>"%(
             self.taskid, self.task.name, self.args, self.kwargs)
 
-class _EnqueuedTask(object):
+class EnqueuedTask(object):
     '''
-    An object that allows for simple status checks on tasks to be executed.
+    An object that allows for simple status checks on tasks being executed right
+    now. Returned by task.execute() or call_task().
     '''
     __slots__ = 'name', 'taskid', 'queue', 'task'
     def __init__(self, name, taskid, queue, task=None):
@@ -1136,8 +1265,14 @@ class _EnqueuedTask(object):
 
             exp = self.task.save_results if self.task and self.task.save_results > 0 else 60
             k = RESULT_KEY + tid
-            if pipe.expire(k, exp).get(k).execute()[-1] == '':
+            v = pipe.expire(k, exp).get(k).execute()[-1]
+            if v in (b'', u''):
                 return "started"
+            # Technically could be 'done' or 'started', but we don't know
+            # because the key that kept that status doesn't exist anymore. Call
+            # it done.
+            # if not v:
+            #     return "unknown"
 
             return "done"
         elif eta <= time.time():
@@ -1157,8 +1292,11 @@ class _EnqueuedTask(object):
 
     def wait(self, timeout=30):
         '''
-        Will wait up to the specified timeout for the task to start execution,
-        returning whether the task started execution.
+        Will wait up to the specified timeout in seconds for the task to start
+        execution, returning True if the task has at least started. Requires
+        task.save_results > 0, and you need to check before the results would
+        have expired for consistent behavior (remember: we can't check status on
+        data that no longer exists).
         '''
         end = time.time() + max(timeout, .01)
         while self.status != 'done' and end > time.time():
@@ -1380,8 +1518,10 @@ def execute_tasks(queues=None, threads_per_process=1, processes=1, wait_per_thre
         pp.daemon = True
         pp.start()
         sp.append(pp)
+    last = {}
     while not SHOULD_QUIT[0]:
         _handle_delayed(get_connection(), queues, 1)
+        last = _update_redis_task_registry(last, get_connection())
 
     log_handler.info("Waiting for %i subprocesses to shutdown", len(sp))
     # wait some time for processes to die...
@@ -1552,7 +1692,7 @@ def clear_queue(queue, conn=None, delete=False, is_data=False):
     if delete:
         pipeline.srem(QUEUES_KNOWN, queue)
         pipeline.zrem(QUEUES_PRIORITY, queue)
-    return pipeline.execute()[0]
+    return sum(pipeline.execute())
 
 def get_task(name):
     '''Get a task dynamically by name. The task's module must be loaded first.'''
